from argaeus.display.aelement import AElement
from PIL import ImageFont
from PIL import ImageDraw
import os


class StaticText:
    _font = None
    _size = None
    _string = None

    def __init__(self, config, verbose):
        AElement.__init__(self, config, verbose)

        self._font = os.path.expanduser(self._config["font"])
        self._size = self._config["size"]
        self._image_font = ImageFont.truetype(self._font, self._size)
        self._string = config["string"]

    def start(self):
        if self._verbose:
            print("StaticText.start()")
        pass

    def stop(self):
        if self._verbose:
            print("StaticText.stop()")
        pass

    def update_image(self):
        if self._verbose:
            print("StaticText.updateImage()")
        # clear image
        draw = ImageDraw.Draw(self.img)
        draw.rectangle((0, 0, self._width, self._height), fill=self._background_color)
        # write time
        if self._verbose:
            print("SetPointText.updateImage() - string '{}'.".format(self._string))
        w, h = draw.textsize(self._string, self._image_font)
        x = int((self._width - w) / 2)
        y = int((self._height - h) / 2)
        draw.text((x, y), self._string, font=self._image_font, fill=self._foreground_color)




from PIL import Image
import os


class AMode:
    _config = None
    _verbose = None

    name = None
    selectable = None
    icon = None

    def __init__(self, config, verbose):
        self._config = config
        self._verbose = verbose
        self.name = self._config["name"]
        if self._verbose:
            print("{}.__init__ - initializing {} ('{}').".format(self.__class__.__name__, self.name, self._config))

        self.selectable = bool(self._config["selectable"])
        icon_path = os.path.expanduser(self._config["icon"])
        self.icon = Image.open(icon_path)



import paho.mqtt.client as mqtt
from threading import Event, Lock
import time


class MQTTClient(object):
    """Wrapper for the paho.mqtt.client. Used by ADriver and DeviceManager but can also be used for other purposes."""

    client = None  # holds an instance of paho.mqtt.client.Client
    _config = None  # json configuration
    _verbose_mqttclient = False  # print debugging information if set to yes.
    is_connected = None  # threading.Event - True if connection to mqtt broker has been successfully established.
    is_disconnected = None
    _topic_handler = None
    _lock_client = None

    def __init__(self, config, verbose):
        self._config = config
        self._verbose_mqttclient = verbose
        self.is_connected = Event()
        self.is_connected.clear()
        self.is_disconnected = Event()
        self.is_disconnected.set()
        self.client = mqtt.Client()
        self._topic_handler = {}
        self._lock_client = Lock()

    def connect(self):
        """Connect to the mqtt broker using the provided configuration and on_message function."""
        if self._verbose_mqttclient:
            print("MQTTClient.connect() - Connecting to mqtt.")
        with self._lock_client:
            self.client.on_connect = self._on_connect
            self.client.on_message = self._on_message
            self.client.username_pw_set(self._config["mqtt-user"], password=self._config["mqtt-password"])
            self.client.connect(self._config["mqtt-address"], self._config["mqtt-port"], 60)
            self.client.loop_start()
            if self.is_connected.wait(30):
                for topic in self._topic_handler.keys():
                    if self._verbose_mqttclient:
                        print("MQTTClient.connect() - subscribe to topic '{}'.".format(topic))
                    self.client.subscribe(topic)
            else:
                raise RuntimeError("MQTTClient.connect - connection to broker could not be established.")

    def disconnect(self):
        """Disconnect from mqtt broker and set is_connected to False."""
        with self._lock_client:
            self.client.disconnect()
            self.is_connected.clear()
            self.is_disconnected.set()

    def _on_connect(self, client, userdata, flags, rc):
        """Return code after trying to connect to mqtt brokder. If successfully connected, is_connected is True."""
        if self._verbose_mqttclient:
            print("MQTTClient._on_connect - Connected with result code " + str(rc))
        if rc == 0:
            self.is_connected.set()
            self.is_disconnected.clear()

    def _on_message(self, client, userdata, msg):
        t = time.time()

        if self._verbose_mqttclient:
            print("MQTTClient._on_message - received message '{}' on topic '{}' @{}s.".
                  format(msg.payload, msg.topic, t))

        for handler in self._topic_handler[msg.topic]:
            if self._verbose_mqttclient:
                print("MQTTClient._on_message - calling handler '{}' ({}).".format(handler, t))
            handler(msg.payload)

    def publish(self, topic, msg):
        if self._verbose_mqttclient:
            print("MQTTClient.publish - publishing to topic '{}' the message '{}'.".format(topic, msg))
        if self.is_connected.is_set():
            self.client.publish(topic, msg)
        else:
            raise RuntimeWarning("MQTTClient.publish - trying to publish while not being connected to mqtt broker.")

    def subscribe(self, topic, handler):
        if self._verbose_mqttclient:
            print("MQTTClient.subscribe - subscribing topic '{}' with handler '{}'.".format(topic, handler))
        with self._lock_client:
            try:
                h = self._topic_handler[topic]
                h.append(handler)
                if h.count(handler) > 1:
                    raise ValueError("MQTTClient.subscribe - topic/handler pair already added. ({}/{})".
                                     format(topic, handler))
            except KeyError:
                self._topic_handler[topic] = [handler,]
            if self.is_connected.is_set():
                if self._verbose_mqttclient:
                    print("MQTTClient.subscribe - activating topic subscription.")
                self.client.subscribe(topic)

    def unsubscribe(self, topic, handler):
        if self._verbose_mqttclient:
            print("MQTTClient.unsubscribe - unsubscribing topic '{}' with handler '{}'.".format(topic, handler))
        with self._lock_client:
            if len(self._topic_handler[topic]) > 1:
                self._topic_handler[topic].remove(handler)
            else:
                del(self._topic_handler[topic])
                self.client.unsubscribe(topic)

    @staticmethod
    def merge(driver, mqtt_client):
        """Utility method - joins a driver instance with an external mqtt-client. Must be called before connecting."""
        driver._mqtt = mqtt_client
        mqtt_client.on_message = driver._on_message


from argaeus.controller.modes.modeprogram import ModeProgram
from argaeus.controller.modes.modeschedule import ModeSchedule
import collections


class ModeFactory:
    @staticmethod
    def create_mode(config, verbose, mqtt_client, config_topics_pub):
        if verbose:
            print("ModeFactory.create_mode - creating mode ('{}').".format(config))
        mode = None

        t = config["type"].lower()
        if t == "program":
            mode = ModeProgram(config, verbose, mqtt_client, config_topics_pub)
        elif t == "schedule":
            mode = ModeSchedule(config, verbose)
        else:
            raise ValueError("ModeFactory.create_mode - unknown type '{}'.".format(t))

        return mode

    @staticmethod
    def create_modes(config, verbose, mqtt_client, config_topics_pub):
        if verbose:
            print("ModeFactory.create_modes - creating modes ('{}').".format(config))

        modes = {}
        modes_selectable = []
        for c in config:
            mode = ModeFactory.create_mode(c, verbose, mqtt_client, config_topics_pub)
            modes[mode.name] = mode
            if mode.selectable:
                modes_selectable.append(mode)
        if verbose:
            print("ModeFactory.create_modes - map_schedule_modes")
        for mode in modes.values():
            if isinstance(mode, ModeSchedule):
                mode.map_schedule_modes(modes)

        return modes, modes_selectable



import datetime
from argaeus.controller.modes.modefactory import ModeFactory
from argaeus.controller.modes.modeschedule import ModeSchedule
from argaeus.controller.acontroller import AController
from argaeus.controller.behavior.buttonbehavior import ButtonBehavior
from argaeus.controller.behavior.behaviorworker import BehaviorWorker


class ModeController(AController):
    _state = None
    _mqtt_client = None

    _modes = None
    _selectable_modes = None
    _selectable_pos = None
    _default_mode = None

    _topic_sub_left = None
    _command_left = None
    _topic_sub_right = None
    _command_right = None

    def __init__(self, config, verbose, state, mqtt_client, config_modes, config_topics_sub, config_topics_pub,
                 config_mqtt_translations):
        AController.__init__(self, config, verbose, config_topics_sub, config_topics_pub, config_mqtt_translations)
        self._state = state
        self._mqtt_client = mqtt_client

        self._modes, self._selectable_modes = ModeFactory.create_modes(config_modes, self._verbose, mqtt_client,
                                                                       config_topics_pub)
        self._default_mode = self._modes[self._config["default-mode"]]
        self._activate_default_mode()

        self._topic_sub_left = self._get_topic_sub(self._config["topic-sub-left"])
        self._command_left = self._get_mqtt_translations(self._config["command-left"])
        self._topic_sub_right = self._get_topic_sub(self._config["topic-sub-right"])
        self._command_right = self._get_mqtt_translations(self._config["command-right"])

        self._behavior_worker = BehaviorWorker(self._verbose, self._state, ButtonBehavior.ToDefaultMode,
                                               self._worker_executor)
        if self._verbose:
            print("ModeController.__init__ - done")

    def _activate_default_mode(self):
        self._selectable_pos = self._selectable_modes.index(self._default_mode)
        self._state.current_mode = self._default_mode
        if isinstance(self._state.current_mode, ModeSchedule):
            dt = datetime.datetime.time(datetime.datetime.now())
            self._state.current_program = self._state.current_mode.get_program_at_time(dt)
        else:
            self._state.current_program = self._state.current_mode

    def _worker_executor(self, item):
        if self._verbose:
            print("ModeController._worker_executor - activate default mode")
        self._activate_default_mode()
        self.update()

    def _topic_handler_left(self, value):
        if value.decode("utf-8") == self._command_left:
            if self._verbose:
                print("ModeController._topic_handler - command left.")
            self._selectable_pos = self._selectable_pos - 1
            self._post_topic_handler()

    def _topic_handler_right(self, value):
        if value.decode("utf-8") == self._command_right:
            if self._verbose:
                print("ModeController._topic_handler - command right.")
            self._selectable_pos = self._selectable_pos + 1
            self._post_topic_handler()

    def _post_topic_handler(self):
        self._selectable_pos = self._selectable_pos % len(self._selectable_modes)
        self._state.current_mode = self._selectable_modes[self._selectable_pos]
        if self._verbose:
            print("ModeController._topic_handler - selected mode '{}' at pos '{}'.".
                  format(self._state.current_mode.name, self._selectable_pos))
        self.update()

    def update(self):
        if isinstance(self._state.current_mode, ModeSchedule):
            dt = datetime.datetime.time(datetime.datetime.now())
            self._state.current_program = self._state.current_mode.get_program_at_time(dt)
            self._state.current_schedule = self._state.current_mode.schedule_raw
        else:
            self._state.current_program = self._state.current_mode
            self._state.current_schedule = None
        self._state.current_program.publish()
        self._state.update_display.set()

    def start(self):
        if self._verbose:
            print("ModeController.start - starting")
        self._mqtt_client.subscribe(self._topic_sub_left, self._topic_handler_left)
        self._mqtt_client.subscribe(self._topic_sub_right, self._topic_handler_right)
        self._state.current_program.publish()
        self._behavior_worker.start()
        self.update()
        if self._verbose:
            print("ModeController.start - started")

    def stop(self):
        if self._verbose:
            print("ModeController.start - stopping")
        self._mqtt_client.unsubscribe(self._topic_sub_left, self._topic_handler_left)
        self._mqtt_client.unsubscribe(self._topic_sub_right, self._topic_handler_right)
        self._behavior_worker.stop()
        if self._verbose:
            print("ModeController.start - stopped")




from argaeus.controller.acontroller import AController
from argaeus.controller.behavior.buttonbehavior import ButtonBehavior
from argaeus.controller.behavior.behaviorworker import BehaviorWorker


class SetPointController(AController):
    _state = None
    _mqtt_client = None

    _step_size = None
    _min_temp = None
    _max_temp = None

    _topic_sub_left = None
    _command_left = None
    _topic_sub_right = None
    _command_right = None

    _behavior_worker = None

    def __init__(self, config, verbose, state, mqtt_client, config_topics_sub, config_mqtt_translations):
        AController.__init__(self, config, verbose, config_topics_sub, None, config_mqtt_translations)
        self._state = state
        self._mqtt_client = mqtt_client

        self._step_size = float(self._config["step-size"])
        self._min_temp = float(self._config["min-temp"])
        self._max_temp = float(self._config["max-temp"])

        self._topic_sub_left = self._get_topic_sub(self._config["topic-sub-left"])
        self._command_left = self._get_mqtt_translations(self._config["command-left"])
        self._topic_sub_right = self._get_topic_sub(self._config["topic-sub-right"])
        self._command_right = self._get_mqtt_translations(self._config["command-right"])

        self._behavior_worker = BehaviorWorker(self._verbose, self._state, ButtonBehavior.ResetToDefaultTemp,
                                               self._worker_executor)
        if self._verbose:
            print("SetPointController.__init__ - done")

    def _worker_executor(self, item):
        if self._verbose:
            print("SetPointController._worker_executor - reset to default set point.")
        self._post_topic_handler(self._state.current_program.default_set_point)

    def _topic_handler_left(self, value):
        if value.decode("utf-8")  == self._command_left:
            if self._verbose:
                print("SetPointController._topic_handler - command left.")
            set_point = self._state.current_program.set_point - self._step_size
            self._post_topic_handler(set_point)

    def _topic_handler_right(self, value):
        if value.decode("utf-8")  == self._command_right:
            if self._verbose:
                print("SetPointController._topic_handler - command right.")
            set_point = self._state.current_program.set_point + self._step_size
            self._post_topic_handler(set_point)

    def _post_topic_handler(self, set_to):

        set_point = min(self._max_temp, max(self._min_temp, set_to))
        if self._verbose:
            print("SetPointController._topic_handler - set temp to '{}'.".format(self._state.current_mode.name))
        self._state.current_program.set_point = set_point
        self._state.current_program.publish()
        self._state.update_display.set()

    def start(self):
        self._mqtt_client.subscribe(self._topic_sub_left, self._topic_handler_left)
        self._mqtt_client.subscribe(self._topic_sub_right, self._topic_handler_right)
        self._behavior_worker.start()

    def stop(self):
        self._mqtt_client.unsubscribe(self._topic_sub_left, self._topic_handler_left)
        self._mqtt_client.unsubscribe(self._topic_sub_right, self._topic_handler_right)
        self._behavior_worker.stop()



import collections
import datetime
from argaeus.controller.modes.amode import AMode


class ModeSchedule(AMode):
    schedule_raw = None
    schedule = None

    def __init__(self, config, verbose):
        AMode.__init__(self, config, verbose)
        self.schedule_raw = ModeSchedule._sort_dict(self._config["schedule"])
        self.schedule = collections.OrderedDict()

    @staticmethod
    def _sort_dict(d):
        result = collections.OrderedDict()
        for h in range (0, 24):
            for m in range(0, 60):
                key = "{:02}:{:02}".format(h, m)
                dt = datetime.time(hour=h, minute=m)
                try:
                    result[dt] = d[key]
                except KeyError:
                    pass
        return result

    def map_schedule_modes(self, modes):
        if self._verbose:
            print("ModeSchedule.map_schedule_modes - mapping schedule '{}' to mode '{}'.".
                  format(self.schedule_raw, modes))
        for k,v in self.schedule_raw.items():
            program = modes[v]
            if isinstance(program, ModeSchedule):
                raise ValueError("ModeSchedule.map_schedule_modes -  mode must not be of type ModeSchedule ('{}').".
                                 format(type(program)))
            self.schedule[k] = program

    def get_program_at_time(self, time):  # time as datetime.time instance
        result = None
        for dt, program in self.schedule.items():
            result = program
            if dt > time:
                break
        if self._verbose:
            print("ModeSchedule.get_program_at_time - program '{}' at time '{}'.".format(result.name, time))
        return result



from argaeus.controller.acontroller import AController
from argaeus.controller.behavior.behaviorworker import BehaviorWorker
from argaeus.controller.behavior.buttonbehavior import ButtonBehavior


class OperationController(AController):
    _mqtt_client = None
    _state = None

    _default_is_active = None
    
    _topic_pub = None
    _command_active = None
    _command_passive = None

    def __init__(self, config, verbose, state, mqtt_client, config_topics_pub, config_mqtt_translations):
        AController.__init__(self, config, verbose, None, config_topics_pub, config_mqtt_translations)
        self._mqtt_client = mqtt_client
        self._state = state

        self._default_is_active = bool(self._config["default-is-active"])
        self._state.active_operation = self._default_is_active

        self._topic_pub = self._get_topic_pub(self._config["topic-pub"])
        self._command_active = self._get_mqtt_translations(self._config["command-active"])
        self._command_passive = self._get_mqtt_translations(self._config["command-passive"])

        self._behavior_worker = BehaviorWorker(self._verbose, self._state, ButtonBehavior.ToggleActivePassive,
                                               self._worker_executor)
        if self._verbose:
            print("OperationController.__init__ - done")

    def _worker_executor(self, item):
        self._state.active_operation = not self._state.active_operation
        if self._verbose:
            print("OperationController._worker_executor - toggle active/passive (now: '{}').".
                  format(self._state.active_operation))
        self._publish()

    def _publish(self):
        if self._state.active_operation:
            self._mqtt_client.publish(self._topic_pub, self._command_active)
        else:
            self._mqtt_client.publish(self._topic_pub, self._command_passive)

    def start(self):
        self._publish()
        self._behavior_worker.start()

    def stop(self):
        self._behavior_worker.stop()



from PIL import Image
import threading
import time
import os
from argaeus.tools.epapermqttmessageconverter import EPaperMQTTMessageEncoder


class Display:
    _config = None
    _verbose = None
    _elements = None

    _width = None
    _height = None

    _lock_update = None

    _mqtt_client = None

    last_image = None
    _base_image = None
    _background_image = None

    _state = None
    _handle_for_update_display_events = None
    _stop_loop = None

    _wipe_screen_at_startup = None

    def __init__(self, config, verbose, state, elements, mqtt_client, topics_pub):
        self._verbose = verbose
        self._config = config
        self._state = state
        if self._verbose:
            print("Display.__init__ - creating instance ('{}', '{}').".format(config, topics_pub))
        self._elements = elements
        self._lock_update = threading.Lock()
        self._mqtt_client = mqtt_client

        self._topic_full_image = topics_pub["epaper"]["full_image"]
        self._topic_full_image_twice = topics_pub["epaper"]["full_image_twice"]
        self._topic_partial_image = topics_pub["epaper"]["part_image"]
        self._topic_partial_image_twice = topics_pub["epaper"]["part_image_twice"]

        self._width = int(self._config["width"])
        self._height = int(self._config["height"])
        self._background_color = int(self._config["background-color"])
        if self._background_color not in [0,255]:
            raise ValueError("Display.__init__ - background-color must be 0 or 255 ('{}').".
                             format(self._background_color))

        self._background_image = Image.open(os.path.expanduser(self._config["background"]))
        self._base_image = Image.new('1', (self._width, self._height), self._background_color)
        self._base_image.paste(self._background_image, (0,0))
        self.last_image = self._base_image.copy()

        self._wipe_screen_at_startup = bool(self._config["wipe-screen"]["at-startup"])
        self._handle_for_update_display_events = threading.Thread(target=self._update_display_thread)
        self._stop_loop = threading.Event()
        self._stop_loop.clear()

        if self._verbose:
            print("Display.__init__ - done.")

    def _update_display_thread(self):
        while not self._stop_loop.isSet():
            if self._state.update_display.wait(0.5):
                self.update()

    def update(self):
        if self._verbose:
            print("Display.update - update display ... waiting for lock")
        with self._lock_update:
            self._state.update_display.clear()  # maybe not necessary to be called - racingcondition between
                                                 # time event and update display event could occoure ...
            self._update_elements()
            new_image = self._merge_elements_to_new_image()
            self._update_display(new_image)
        if self._verbose:
            print("Display.update - update display ... released lock")

    def _update_elements(self):
        if self._verbose:
            print("Display._update_elements - update display")
        for element in self._elements:
            try:
                element.update_image()
            except NotImplementedError:
                pass

    def _merge_elements_to_new_image(self):
        if self._verbose:
            print("Display._merge_elements - create new image with the display")
        new_image = self._base_image.copy()
        for element in self._elements:
            new_image.paste(element.img, (element.x, element.y))
        return new_image

    def _update_display(self, new_image):
        message = EPaperMQTTMessageEncoder.to_full_image_message(new_image)
        self._mqtt_client.publish(self._topic_full_image, message)
        self.last_image = new_image

    def _wipe_screen(self):
        if self._verbose:
            print("Display._wipe_screen - begin")
        with self._lock_update:
            img_white = Image.new('1', (self._width, self._height), 255)
            img_black = Image.new('1', (self._width, self._height), 255)
            msg_white = EPaperMQTTMessageEncoder.to_full_image_message(img_white)
            msg_black = EPaperMQTTMessageEncoder.to_full_image_message(img_black)
            for x in range(4):
                if self._verbose:
                    print("Display._wipe_screen - sending black&white combination (cycle {}).".format(x))
                self._mqtt_client.publish(self._topic_full_image_twice, msg_black)
                time.sleep(5)
                self._mqtt_client.publish(self._topic_full_image_twice, msg_white)
                time.sleep(5)
        if self._verbose:
            print("Display._wipe_screen - end")

    def start(self, skip_wipe_screen=False):  # skip_wipe_screen for testing purposes
        if self._verbose:
            print("Display.start")
        if self._wipe_screen_at_startup and not skip_wipe_screen:
            self._wipe_screen()
        if self._verbose:
            print("Display.start - starting {} elements.".format(len(self._elements)))
        for e in self._elements:
            try:
                e.start()
            except NotImplementedError:
                pass
        self._stop_loop.clear()
        self._handle_for_update_display_events.start()

    def stop(self):
        if self._verbose:
            print("Display.stop - stopping {} elements.".format(len(self._elements)))
        for e in self._elements:
            try:
                e.stop()
            except NotImplementedError:
                pass
        self._stop_loop.set()
        self._handle_for_update_display_events.join()



from argaeus.display.aelement import AElement
from PIL import Image
import os


class ModeIcon(AElement):
    _get_current_mode_fct = None
    _state = None

    def __init__(self, config, verbose, state):
        AElement.__init__(self, config, verbose)
        self._state = state

    def start(self):
        if self._verbose:
            print("ModeIcon.start()")
        pass

    def stop(self):
        if self._verbose:
            print("ModeIcon.stop()")
        pass

    def update_image(self):
        icon = self._state.current_mode.icon
        w,h = icon.size
        x = int((self._width - w) / 2)
        y = int((self._height - h) / 2)
        self.img = Image.new('1', (self._width, self._height), self._background_color)
        self.img.paste(icon, (x, y))



from argaeus.display.aelement import AElement
from PIL import ImageDraw


class Schedule(AElement):
    _config_schedule = None
    _patterns = None
    _pattern_schedule = None
    _group_by = None  # in minutes. 60 must be a multiple of this value -> 1, 2, 3, 4, 5, 6, 10, 12, 15, 20, 30, 60
    _pixel_per_slot = None  # x-axis
    _pixel_per_dot = None  # y-axis
    _state = None

    def __init__(self, config, verbose, state):
        AElement.__init__(self, config, verbose)
        self._state = state

        self._group_by = int(self._config["group-by"])
        if 60 % self._group_by != 0:
            raise ValueError("Schedule - 60 must be a multiple of group-by. (60 % '{}' != 0)".
                             format(self._group_by))
        groups_per_hour = int(60 / self._group_by)
        pixel_per_hour = int(self._width / 24)
        if pixel_per_hour % groups_per_hour != 0:
            raise ValueError("Schedule - width ({}) and group-by ({}) not fulfilling constraint "
                             "'(width / 24) % (60 / group-by) == 0'".format(self._width, self._group_by))
        self._pixel_per_slot = int(pixel_per_hour / groups_per_hour)

        if self._height % 2 != 0:
            raise ValueError("Schedule - height ({}) must be dividable by 2.".format(self._height))
        self._pixel_per_dot = int(self._height / 2)

        self._patterns = self._config["patterns"]
        self._pattern_schedule = []

    def _get_pattern_schedule(self):
        result = []
        try:
            schedule_items = self._state.current_schedule.items()
            if self._verbose:
                print("Schedule._fill_pattern_schedule - currently active schedule '{}'. (patterns: '{}')".
                      format(self._state.current_schedule, self._patterns))
        except AttributeError:
            if self._verbose:
                print("Schedule._fill_pattern_schedule - currently no schedule active.")
        else:
            for k,v in schedule_items:
                p = self._patterns[v.lower()]
                if self._verbose:
                    print("Schedule._fill_pattern_schedule - [{}]: add pattern {}/{}.".
                          format(k, v, p))
                result.append(p)
        return result

    def start(self):
        if self._verbose:
            print("Schedule.start()")

    def stop(self):
        if self._verbose:
            print("Schedule.stop()")

    def update_image(self):
        if self._verbose:
            print("Schedule.updateImage() - pixel/dot: {}, pixel/slot: {}.".
                  format(self._pixel_per_dot, self._pixel_per_slot))

        pattern_schedule = self._get_pattern_schedule()

        # clear image
        draw = ImageDraw.Draw(self.img)
        draw.rectangle((0, 0, self._width, self._height), fill=self._background_color)
        # draw graph
        count = 0
        for pattern in pattern_schedule:
            if self._verbose:
                print("Schedule.update_image - {}. pattern is {}.".format(count, pattern))
            x_from = count * self._pixel_per_slot
            x_to = x_from + self._pixel_per_slot - 1
            count = count + 1
            if pattern == 0: # nothing is displayed
                continue
            elif pattern == 1: # lower dot
                y_lower = self._height - 1
                y_upper = self._height - self._pixel_per_dot
            elif pattern == 2: # full height
                y_lower = self._height - 1
                y_upper = 0
            elif pattern == 3: # upper dot
                y_lower = self._height - self._pixel_per_dot - 1
                y_upper = 0
            else:
                raise ValueError("Schedule.update_image - unknown value for pattern '{}'.".format(pattern))
            if self._verbose:
                print("Schedule.update_image - rectangle({}/{}-{}/{}).".
                      format(count, pattern, x_from, y_lower, x_to, y_upper))
            draw.rectangle((x_from, y_lower, x_to, y_upper), fill=self._foreground_color)




import datetime
from PIL import ImageDraw
from argaeus.display.aelementmqtt import AElementMQTT
from argaeus.tools.myinfluxdbclient import MyInfluxDBClient
from argaeus.display.graphdata import GraphData


class Graph(AElementMQTT):
    _influx_client = None
    _influx_data_series = None

    _group_by = None  # in minutes. 60 must be a multiple of this value -> 1, 2, 3, 4, 5, 6, 10, 12, 15, 20, 30, 60
    _pixel_per_slot = None
    _min_degree_per_dot = None # with a height of e.g. 12 dots this value means that the graph will span at least 6
                               # degrees vertically. this min_value prevents that very small temp variants in 24 hours
                               # are equally depicted as huge changes in 24 hours. the 6 degrees in this example should
                               # cover most smart home heating situations.
    _epoch = None
    _epoch_size = None  # group_by as time object
    _next_epoch = None

    _value_buffer = None
    _data = None

    def __init__(self, config_graph, config_influxdb, config_topics_sub, verbose, mqtt_client):
        AElementMQTT.__init__(self, config_graph, config_topics_sub, verbose, mqtt_client)

        try:
            self._influx_data_series = str(config_graph["influx-dataseries"])
        except KeyError:
            if self._verbose:
                print("Graph.__init__ - dont use influxdb.")
            self._influx_data_series = ""

        if len(self._influx_data_series) > 0:
            self._influx_client = MyInfluxDBClient(config_influxdb, verbose)

        if self._width % 24 != 0:
            raise ValueError("Graph - width must be a multiple of 24. ('{}' % 24 != 0)".
                             format(self._width))

        self._group_by = int(self._config["group-by"])
        if 60 % self._group_by != 0:
            raise ValueError("Graph - 60 must be a multiple of group-by. (60 % '{}' != 0)".
                             format(self._group_by))
        self._epoch_size = datetime.time(minute=self._group_by)
        self._epoch = datetime.time()
        self._set_next_epoch()

        groups_per_hour = int(60 / self._group_by)
        pixel_per_hour = int(self._width / 24)
        if pixel_per_hour % groups_per_hour != 0:
            raise ValueError("Graph - width ({}) and group-by ({}) not fulfilling constraint "
                             "'(width / 24) % (60 / group-by) == 0'".format(self._width, self._group_by))
        self._pixel_per_slot = int(pixel_per_hour / groups_per_hour)

        self._min_degree_per_dot = float(self._config["min_degree_per_dot"])

        self._value_buffer = []
        self._data = GraphData(self._verbose, groups_per_hour)

        if self._verbose:
            print("Graph.__init__ - initialized data('{}')".format(self._data.slots.keys()))

    def _set_next_epoch(self):
        current_dt = datetime.datetime.combine(datetime.date(1, 1, 1), self._epoch)
        next_dt = current_dt + datetime.timedelta(minutes=self._epoch_size.minute)
        self._next_epoch = next_dt.time()
        if self._verbose:
            print("Graph._set_next_epoch - next epoch starts at '{}'. current started at '{}'.".
                  format(self._next_epoch, self._epoch))

    def _get_history(self):
        """fill all slots with data from influxdb."""
        if self._verbose:
            print("Graph._get_history - fetching history from influxdb.")

        result = []
        query = "select mean(*) from {} where time>now()-1d group by time({}m)".format(self._influx_data_series,
                                                                                      self._group_by)
        if self._verbose:
            print("Graph._get_history - query-string: '{}'.".format(query))
        query_result = self._influx_client.client.query(query)
        if self._verbose:
            print("Graph._get_history - query-result: '{}'.".format(query_result))
        self._data.fill_history(query_result, "mean_value")
        if self._verbose:
            print("Graph._get_history - slots: '{}'.".format(self._data.slots))

    def _topic_sub_handler(self, value, t=None):  # t for debugging
        if t is None:
            t = datetime.datetime.time(datetime.datetime.now())
        msg_epoch = self._time_to_epoch(t)
        self._value_buffer.append({"time": t, "value": float(value)})
        if self._verbose:
            print("Graph._topic_sub_handle - received value '{}' for epoch '{}'.".format(value, msg_epoch))

        if (msg_epoch >= self._next_epoch) or (msg_epoch < self._epoch):
            # new epoch. second part of if means that current epoch is the last of the day and new epoch is in hour '0'.
            self._set_current_epoch_slot_value()
            if self._verbose:
                print("Graph._topic_sub_handle - new epoch '{}' (old: '{}').".format(msg_epoch, self._epoch))
            self._epoch = msg_epoch  # forward current epoch to msg_epoch
            self._set_next_epoch()

    def _set_current_epoch_slot_value(self):
        value_sum = 0
        count = 0
        if self._verbose:
            print("Graph._set_current_epoch_slot_value - current epoch: '{}'. next epoch starts at '{}'. '{}' values "
                  "in buffer ('{}').".format(self._epoch, self._next_epoch, len(self._value_buffer),
                                             self._value_buffer))
        while True:
            if len(self._value_buffer) == 0:
                if self._verbose:
                    print("Graph._set_current_epoch_slot_value - empty buffer")
                break
            if self._value_buffer[0]["time"] >= self._next_epoch:
                if self._verbose:
                    print("Graph._set_current_epoch_slot_value - buffer contains no more values for this epoch.")
                break
            entry = self._value_buffer.pop(0)
            if self._verbose:
                print("Graph._set_current_epoch_slot_value - popped entry '{}'.".format(entry))
            value_sum = value_sum + entry["value"]
            count = count + 1
        try:
            avg = value_sum / count
        except ZeroDivisionError:
            avg = 0
        self._data.slots[self._epoch] = avg
        if self._verbose:
            print("Graph._set_current_epoch_slot_value - [{}]='{}' (count={}).".format(self._epoch, avg, count))

    def _time_to_epoch(self, t):
        """convert given time to the time object representing its epoch."""
        m = t.minute - (t.minute % self._group_by)
        e = datetime.time(hour=t.hour, minute=m)
        if self._verbose:
            print("Graph._time_to_epoch - {}, {}, {}.".format(t, m, e))
        return e

    def start(self):
        if self._verbose:
            print("Graph.start()")
        self._epoch = self._time_to_epoch(datetime.datetime.time(datetime.datetime.now()))
        if self._influx_client:
            self._get_history()
        pass

    def stop(self):
        if self._verbose:
            print("Graph.stop()")
        pass

    def update_image(self):
        if self._verbose:
            print("Graph.updateImage()")

        min_value, max_value = self._data.get_min_max_values()
        value_per_dot = (max_value - min_value) / (self._height - 1)
        value_per_dot = max(value_per_dot, self._min_degree_per_dot)

        if self._verbose:
            print("Graph.updateImage - min_value: {}, max_value: {}, value_per_dot: {}.".
                  format(min_value, max_value, value_per_dot))

        # clear image
        draw = ImageDraw.Draw(self.img)
        draw.rectangle((0, 0, self._width, self._height), fill=self._background_color)

        # draw graph
        count = 0
        for value in self._data.slots.values():
            x = count * self._pixel_per_slot
            y = self._height - round((value - min_value) / value_per_dot) - 1
            if y < 0:  # rounding may lead to values that are out of bound
                y = 0
            if y >= self._height:  # rounding may lead to values that are out of bound
                y = self._height - 1
            if self._verbose:
                print("Graph.updateImage - value: {}, x: {}, y: {}.".format(value, x, y))
            draw.line(((x, y), (x + self._pixel_per_slot - 1, y)), fill=self._foreground_color)
            count = count + 1

        # draw vertical line for the ongoing epoch
        current_slot = int(self._epoch.hour * 60 + self._epoch.minute) / int(self._epoch_size.minute)
        x = current_slot * self._pixel_per_slot
        if self._verbose:
            print("Graph.updateImage - vertical line: epoch: ({}h, {}m, {}size), slot: {}, x: {}.".
                  format(self._epoch.hour, self._epoch.minute, self._epoch_size, current_slot, x))
        draw.rectangle((x, 0, x + self._pixel_per_slot, self._height), fill=self._foreground_color)




from argaeus.controller.modes.amode import AMode


class ModeProgram(AMode):
    set_point = None
    default_set_point = None
    _mqtt_client = None
    _topic_pub = None

    def __init__(self, config, verbose, mqtt_client, config_topics_pub):
        AMode.__init__(self, config, verbose)
        self.set_point = float(self._config["set-point"])
        self.default_set_point = self.set_point
        self._mqtt_client = mqtt_client
        self._topic_pub = config_topics_pub["temperature"]["set-point"]

    def publish(self):
        self._mqtt_client.publish(self._topic_pub, self.set_point)




import threading
import queue
from argaeus.controller.behavior.buttonbehavior import ButtonBehavior


class State:
    active_operation = None
    current_mode = None
    current_program = None
    current_schedule = None
    update_display = None
    behavior_queues = None

    def __init__(self):
        self.update_display = threading.Event()
        self.update_display.clear()

        self.behavior_queues = {}
        for b in ButtonBehavior:
            self.behavior_queues[b] = queue.Queue()




from influxdb import InfluxDBClient

class MyInfluxDBClient:
    client = None
    _influxdb_username = None
    _influxdb_password = None
    _influxdb_address = None
    _influxdb_port = None
    _influxdb_database = None

    _config = None
    _verbose = None

    def __init__(self, config, verbose):
        self._config = config
        self._verbose = verbose
        if self._verbose:
            print("MyInfluxDBClient.__init__ - creating instance ('{}').".format(self._config))

        self._influxdb_address = str(self._config["influx-address"])
        self._influxdb_port = int(self._config["influx-port"])
        self._influxdb_database = str(self._config["database"])
        self._influxdb_password = str(self._config["influx-password"])
        self._influxdb_username = str(self._config["influx-user"])

        self.client = InfluxDBClient(host=self._influxdb_address, port=self._influxdb_port,
                                       username=self._influxdb_username, password=self._influxdb_password,
                                       database=self._influxdb_database)



from argaeus.display.aelement import AElement


class AElementMQTT(AElement):
    _mqtt_client = None
    _topic_sub = None

    def __init__(self, config, config_topics_sub, verbose, mqtt_client):
        AElement.__init__(self, config, verbose)
        self._mqtt_client = mqtt_client
        group, element = self._config["topic-sub"].split(".")
        self._topic_sub = config_topics_sub[group][element]
        self._mqtt_client.subscribe(self._topic_sub, self._topic_sub_handler)

    def _topic_sub_handler(self, value):
        raise NotImplementedError



from argaeus.display.graph import Graph
from argaeus.display.schedule import Schedule
from argaeus.display.modeicon import ModeIcon
from argaeus.display.digitalclock import DigitalClock
from argaeus.display.mqtttext import MQTTText
from argaeus.display.setpointtext import SetPointText
from argaeus.display.statictext import StaticText
from argaeus.display.bar import Bar


class ElementFactory:
    @staticmethod
    def create_element(config_element, verbose, mqtt_client, config, state):
        element = None
        if config_element["active"]:
            if config_element["type"].lower() == "graph":
                element = Graph(config_element, config["influx"], config["topics-sub"], verbose, mqtt_client)
            elif config_element["type"].lower() == "schedule":
                element = Schedule(config_element, verbose, state)
            elif config_element["type"].lower() == "modeicon":
                element = ModeIcon(config_element, verbose, state)
            elif config_element["type"].lower() == "setpointtext":
                element = SetPointText(config_element, verbose, state)
            elif config_element["type"].lower() == "statictext":
                element = StaticText(config_element, verbose)
            elif config_element["type"].lower() == "digitalclock":
                element = DigitalClock(config_element, verbose)
            elif config_element["type"].lower() == "mqtttext":
                element = MQTTText(config_element, config["topics-sub"], verbose, mqtt_client)
            elif config_element["type"].lower() == "bar":
                element = Bar(config_element, config["topics-sub"], verbose, mqtt_client)
            else:
                raise ValueError("ElementFactory.create_element - unknown type '{}'".
                                 format(config_element["type"].lower()))
        else:
            if verbose:
                print("ElementFactory.create_element - skipping inactive element '{}.{}'.".
                      format(config_element["type"].lower(), config_element["name"]))

        return element

    @staticmethod
    def create_elements(config_elements, verbose, mqtt_client=None, config=None, state=None):
        element_list = []

        if verbose:
            print("ElementFactory.create_elements - start")

        for config_element in config_elements:
            element = ElementFactory.create_element(config_element, verbose, mqtt_client, config, state)
            if element is not None:
                element_list.append(element)

        if verbose:
            print("ElementFactory.create_elements - finished")

        return element_list





import collections
import datetime


class GraphData:
    _verbose = None
    slots = None
    MIN = -65536
    MAX = 65536

    def __init__(self, verbose, groups_per_hour):
        self._verbose = verbose
        self.slots = collections.OrderedDict()

        group_by = 60 / groups_per_hour
        for h in range(24):
            for g in range(groups_per_hour):
                key = datetime.time(hour=h, minute=int(g*group_by))
                self.slots[key] = 0

    def fill_history(self, history, entry_name):
        for result in history:
            count = 0
            keys = []  # use instead timestamps of query result
            for k in self.slots.keys():
                keys.append(k)
            for entry in result:
                e = keys[count]
                if self._verbose:
                    print(e, entry)
                value = entry[entry_name]
                if value is not None:
                    self.slots[e] = value
                count = count + 1
                if count >= len(self.slots):
                    count = 0

    def get_min_max_values(self):
        min_value = self.MAX
        max_value = self.MIN
        for value in self.slots.values():
            min_value = min(value, min_value)
            max_value = max(value, max_value)
        return min_value, max_value



import argparse
import os
import time
from pathlib import Path
from argaeus.display.display import Display
from argaeus.controller.controllermanager import ControllerManager
from argaeus.display.elementfactory import ElementFactory
from argaeus.state import State
from argaeus.tools import mypyyaml
from argaeus.tools.mqttclient import MQTTClient
from argaeus.tools.myinfluxdbclient import MyInfluxDBClient
import argaeus
import threading


class GUI:
    _config = None
    _verbose = None

    _mqtt_client = None
    _influxdb_client = None

    _state = None
    _display = None
    _controller = None

    _stop_loop = None
    _loop_thread = None
    _is_stopped = None

    def __init__(self, config, verbose):
        self._verbose = verbose
        self._config = config
        if self._verbose:
            print("GUI.__init__ - initializing GUI ('{}').".format(self._config))

        self._mqtt_client = MQTTClient(self._config["mqtt"], self._verbose)
        self._influxdb_client = MyInfluxDBClient(config["influx"], self._verbose)

        self._state = State()
        self._controller = ControllerManager(config, verbose, self._state, self._mqtt_client)

        elements = ElementFactory.create_elements(self._config["display"]["elements"], self._verbose, self._mqtt_client,
                                                  self._config, self._state)
        self._display = Display(self._config["display"], self._verbose, self._state, elements,
                                self._mqtt_client, self._config["topics-pub"])

        self._stop_loop = threading.Event()
        self._is_stopped = threading.Event()
        self._is_stopped.set()
        self._loop_thread = threading.Thread(target=self._poll_loop)

        if self._verbose:
            print("GUI.__init__ - done.")

    def _calc_sleep_time(self):
        safety = 0.5  # wait n additional seconds
        current_time = time.time()
        seconds_to_next_full_minute = (60 - current_time % 60) + safety  # next full minute in seconds
        if self._verbose:
            print("GUI._calc_sleep_time - seconds to next full minute: '{} s' ({}).".
                  format(seconds_to_next_full_minute, current_time))
        return seconds_to_next_full_minute

    def _poll_loop(self):
        if self._verbose:
            print("GUI._poll_loop - entered poll_loop method.")

        while not self._stop_loop.isSet():
            print("update GUI @ {}.".format(time.time()))
            self._controller.update()
            self._display.update()
            sleep_for = self._calc_sleep_time()
            if self._verbose:
                print("GUI._poll_loop - sleep for " + str(sleep_for) + " seconds.")
            self._stop_loop.wait(sleep_for)

        if self._verbose:
            print("GUI._poll_loop - exiting poll_loop method.")

    def start(self):
        if self._verbose:
            print("GUI.start - starting")
        self._is_stopped.clear()
        self._mqtt_client.connect()
        self._mqtt_client.is_connected.wait()
        self._controller.start()
        self._display.start()
        self._loop_thread.start()
        if self._verbose:
            print("GUI.start - started")

    def stop(self):
        if self._verbose:
            print("GUI.stop - stopping")
        self._stop_loop.set()
        self._display.stop()
        self._controller.stop()
        self._loop_thread.join()
        self._mqtt_client.disconnect()
        self._mqtt_client.is_disconnected.wait()
        self._is_stopped.set()
        if self._verbose:
            print("GUI.stop - stopped")

    @classmethod
    def _args_to_config(cls, args=None):
        """Handle command line arguments and read the yaml file into a json structure (=config)."""
        desc = "GUI for thermostat."
        ap = argparse.ArgumentParser(description=desc)
        ap.add_argument('-c', '--config', type=str, help='yaml config file', required=True)
        ap.add_argument('-v', '--verbose', help='verbose', action="store_true")
        ap.add_argument('--version', action='version',
                            version='%(prog)s {}'.format(argaeus.version),
                            help='show the version number and exit')
        if args:
            arguments = vars(ap.parse_args(args))
        else:
            arguments = vars(ap.parse_args())

        verbose = False
        if arguments["verbose"]:
            verbose = True

        config_filename = arguments["config"]
        config_file = Path(config_filename)
        if not config_file.is_file():
            raise FileNotFoundError("config file '{}' not found.".format(config_filename))

        config = mypyyaml.load(open(config_filename, 'r'), Loader=mypyyaml.Loader)

        try:
            credentials_mqtt = mypyyaml.load(open(os.path.expanduser(config["mqtt"]["mqtt-credentials"]), 'r'),
                                        Loader=mypyyaml.Loader)
        except KeyError:
            pass
        else:
            config["mqtt"].update(credentials_mqtt["mqtt"])

        try:
            credentials_influx = mypyyaml.load(open(os.path.expanduser(config["influx"]["influx-credentials"]), 'r'),
                                        Loader=mypyyaml.Loader)
        except KeyError:
            pass
        else:
            config["influx"].update(credentials_influx["influx"])

        return config, verbose

    @classmethod
    def standalone(cls, args=None):
        """Public method to start this driver directly. Instantiates an MQTT client and creates an object for the
                given driver."""
        config, verbose = GUI._args_to_config(args)
        config = mypyyaml.dict_deepcopy_lowercase(config)
        gui = GUI(config, verbose)
        gui.start()
        try:
            while not gui._is_stopped.wait(0.1):  # timeout is necessary for CTRL+C
                pass
        except KeyboardInterrupt:
            pass
        gui.stop()


def standalone():
    GUI.standalone()


if __name__ == "__main__":
    GUI.standalone()



class AController:
    _config = None
    _verbose = None
    _config_topics_sub = None
    _config_topics_pub = None
    _config_mqtt_translations = None

    def __init__(self, config, verbose, config_topics_sub, config_topics_pub, config_mqtt_translations):
        self._config = config
        self._verbose = verbose
        if self._verbose:
            print("{}.__init__ - initializing instance(''{}').".format(self.__class__.__name__, self._config))

        self._config_topics_sub = config_topics_sub
        self._config_topics_pub = config_topics_pub
        self._config_mqtt_translations = config_mqtt_translations

    def _get_topic_sub(self, name):
        group, entry = name.split(".")
        return self._config_topics_sub[group][entry]

    def _get_topic_pub(self, name):
        group, entry = name.split(".")
        return self._config_topics_pub[group][entry]

    def _get_mqtt_translations(self, name):
        group, entry = name.split(".")
        return self._config_mqtt_translations[group][entry]

    def start(self):
        raise NotImplementedError

    def stop(self):
        raise NotImplementedError



import base64
from io import BytesIO
import json


class EPaperMQTTMessageEncoder(object):
    """Static utility class - converts images and json structures into valid mqtt payloads."""

    @staticmethod
    def to_full_image_message(image):
        """Convert a PIL.Image instance to bytes - the format nedded if the mqtt payload consists of only the image."""
        bytes_image = BytesIO()
        image.save(bytes_image, format="png")
        result = bytes_image.getvalue()
        return result

    @staticmethod
    def to_partial_images_message(image_entry_list):
        """Takes a list containing [x,y,partial images] and converts the images into an utf-8 encoded string that can be
        accepted by mqtt and packs them into a json structure consisting of these string and their x/y values."""
        result = []
        for image_entry in image_entry_list:
            bytes_image = BytesIO()
            image_entry["image"].save(bytes_image, format="png")
            base64_bytes = base64.b64encode(bytes_image.getvalue())
            base64_string = base64_bytes.decode("utf-8")
            entry = {
                "x": int(image_entry["x"]),
                "y": int(image_entry["y"]),
                "image": base64_string
            }
            result.append(entry)
        return json.dumps(result)




import math
from argaeus.display.aelementmqtt import AElementMQTT
from PIL import Image


class Bar(AElementMQTT):
    _current_value = None
    _min_value = None
    _max_value = None
    _normalize_to = None

    def __init__(self, config, config_topics_sub, verbose, mqtt_client):
        AElementMQTT.__init__(self, config, config_topics_sub, verbose, mqtt_client)

        self._max_value = float(self._config["max-value"])
        self._min_value = float(self._config["min-value"])
        self._normalize_to = math.log(self._max_value - self._min_value + 1)

        self._current_value = min(self._max_value, max(self._min_value, float(0)))

    def _topic_sub_handler(self, value):
        self._current_value = min(self._max_value, max(self._min_value, float(value)))
        if self._verbose:
            print("Bar._topic_sub_handler - received value '{}' -> current value '{}'.".
                  format(value, self._current_value))

    def start(self):
        if self._verbose:
            print("Bar.start()")
        pass

    def stop(self):
        if self._verbose:
            print("Bar.stop()")
        pass

    def _get_bar_height(self):
        value = self._current_value - self._min_value + 1
        if self._verbose:
            print("Bar._get_bar_height - value: {}, internal value: {}, min: {}, height: {}.".
                  format(self._current_value, value, self._min_value, self._height))
        norm_value = math.log(value) / self._normalize_to
        bar_height = int((self._height - 1) * norm_value) + 1  # +/- 1 sets height of bar to 1 for value 0.
        return bar_height

    def update_image(self):
        bar_height = self._get_bar_height()
        if self._verbose:
            print("Bar.updateImage() - volt '{}', bar height '{}'".
                  format(self._current_value, bar_height))
        # create image
        self.img = Image.new("1", (self._width, bar_height), self._foreground_color)





from PIL import Image


class AElement:
    _config = None
    _verbose = None
    name = None
    img = None
    _width = None
    _height = None
    x = None
    y = None
    _foreground_color = None
    _background_color = None

    def __init__(self, config, verbose):
        self._config = config
        self._verbose = verbose
        if verbose:
            print("{}.__init__ - creating instance ('{}').".format(self.__class__.__name__, config))
        if not self.__class__.__name__.lower() == self._config["type"].lower():
            raise ValueError("type '{}' as given in config is not equvivalent to class type '{}'.".
                             format(self._config["type"].lower(), self.__class__.__name__.lower()))
        self.name = self._config["name"]

        self._width = int(self._config["width"])
        self._height = int(self._config["height"])
        self.x = int(self._config["x"])
        self.y = int(self._config["y"])
        self._foreground_color = int(self._config["foreground-color"])
        self._background_color = int(self._config["background-color"])

        if self._foreground_color not in [0,255]:
            raise ValueError("AElement.'{}' - foreground-color must be 0 or 255 ('{}').".
                             format(self.name, self._foreground_color))
        if self._background_color not in [0,255]:
            raise ValueError("AElement.'{}' - background-color must be 0 or 255 ('{}').".
                             format(self.name, self._background_color))

        self.img = Image.new('1', (self._width, self._height), self._background_color)

    def start(self):
        raise NotImplementedError

    def stop(self):
        raise NotImplementedError

    def update_image(self):
        raise NotImplementedError
